@@ -381,7 +381,7 @@ QT_BEGIN_NAMESPACE
 */
 
 /*!
-    \qmlmethod AbstractBarSeries::clear()
+    \qmlmethod void AbstractBarSeries::clear()
     Removes all bar sets from the series.
 */
 
@@ -515,6 +515,7 @@ bool QAbstractBarSeries::insert(int index, QBarSet *set)
     if (success) {
         QList<QBarSet *> sets;
         sets.append(set);
+        set->setParent(this);
         emit barsetsAdded(sets);
         emit countChanged();
     }
@@ -1047,8 +1048,8 @@ bool QAbstractBarSeriesPrivate::insert(int index, QBarSet *set)
                      this, &QAbstractBarSeriesPrivate::handleSetValueAdd);
     QObject::connect(set->d_ptr.data(), &QBarSetPrivate::valueRemoved,
                      this, &QAbstractBarSeriesPrivate::handleSetValueRemove);
-    disconnect(set, &QBarSet::selectedBarsChanged,
-               this, &QAbstractBarSeriesPrivate::updatedBars);
+    connect(set, &QBarSet::selectedBarsChanged,
+            this, &QAbstractBarSeriesPrivate::updatedBars);
 
     emit restructuredBars();      // this notifies barchartitem
     return true;
