@@ -1,6 +1,8 @@
-// Copyright (C) 2021 The Qt Company Ltd.
+// Copyright (C) 2022 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
+#undef QT_NO_FOREACH // this file contains unported legacy Q_FOREACH uses
+
 #include <QtCharts/QXYSeries>
 #include <QtCharts/QColorAxis>
 #include <private/qxyseries_p.h>
@@ -26,7 +28,7 @@ QT_BEGIN_NAMESPACE
 */
 /*!
     \qmltype XYSeries
-    \instantiates QXYSeries
+    \nativetype QXYSeries
     \inqmlmodule QtCharts
 
     \inherits AbstractSeries
@@ -58,8 +60,15 @@ QT_BEGIN_NAMESPACE
            This enum value can be used to hide or show the label of the point. If used
            together with QXYSeries::setPointConfiguration, the configuration's value
            should be boolean.
+    \value [since 6.5] LabelFormat
+           This enum value can be used to set custom label text per-point. If used together with
+           QXYSeries::setPointConfiguration, the configuration's value should be a string.
+           \note If an empty string is set as the LabelFormat, it will be ignored, and the series
+           pointLabelsFormat will be used.
+           \sa pointLabelsFormat
 
     \sa setPointConfiguration()
+    \since 6.2
 */
 
 /*!
@@ -146,6 +155,8 @@ QT_BEGIN_NAMESPACE
 
     QXYSeries supports the following format tags:
     \table
+      \row
+        \li @index       \li The index in the series of the data point. [since 6.5]
       \row
         \li @xPoint      \li The x-coordinate of the data point.
       \row
@@ -157,7 +168,7 @@ QT_BEGIN_NAMESPACE
     (x, y):
 
     \code
-    series->setPointLabelsFormat("(@xPoint, @yPoint)");
+    series->setPointLabelsFormat("@index: (@xPoint, @yPoint)");
     \endcode
 
     By default, the labels' format is set to \c {@xPoint, @yPoint}. The labels
@@ -307,7 +318,7 @@ QT_BEGIN_NAMESPACE
     \since 6.2
 */
 /*!
-    \qmlproperty bool XYSeries::bestFitLineColor
+    \qmlproperty color XYSeries::bestFitLineColor
     The color of best fit line.
     \sa bestFitLineVisible
 */
@@ -508,6 +519,11 @@ QT_BEGIN_NAMESPACE
     This signal is emitted when the pen changes to \a pen.
 */
 
+/*!
+    \fn void QXYSeries::selectedPointsChanged()
+    This signal is emitted when the points selection changes.
+*/
+
 /*!
     \fn void QXYSeries::lightMarkerChanged(const QImage &lightMarker)
     This signal is emitted when the light marker image changes to \a lightMarker.
@@ -521,36 +537,36 @@ QT_BEGIN_NAMESPACE
 */
 
 /*!
-    \qmlmethod XYSeries::append(real x, real y)
+    \qmlmethod void XYSeries::append(real x, real y)
     Appends a point with the coordinates \a x and \a y to the series.
 */
 
 /*!
-    \qmlmethod XYSeries::replace(real oldX, real oldY, real newX, real newY)
+    \qmlmethod void XYSeries::replace(real oldX, real oldY, real newX, real newY)
     Replaces the point with the coordinates \a oldX and \a oldY with the point
     with the coordinates \a newX and \a newY. Does nothing if the old point does
     not exist.
 */
 
 /*!
-    \qmlmethod XYSeries::remove(real x, real y)
+    \qmlmethod void XYSeries::remove(real x, real y)
     Removes the point with the coordinates \a x and \a y from the series. Does
     nothing if the point does not exist.
 */
 
 /*!
-    \qmlmethod XYSeries::remove(int index)
+    \qmlmethod void XYSeries::remove(int index)
     Removes the point at the position specified by \a index from the series.
 */
 
 /*!
-    \qmlmethod XYSeries::removePoints(int index, int count)
+    \qmlmethod void XYSeries::removePoints(int index, int count)
     Removes the number of points specified by \a count from the series starting
     at the position specified by \a index.
 */
 
 /*!
-    \qmlmethod XYSeries::insert(int index, real x, real y)
+    \qmlmethod void XYSeries::insert(int index, real x, real y)
     Inserts a point with the coordinates \a x and \a y to the position specified
     by \a index in the series. If the index is 0 or less than 0, the point is
     prepended to the list of points. If the index is equal to or greater than
@@ -559,7 +575,7 @@ QT_BEGIN_NAMESPACE
 */
 
 /*!
-    \qmlmethod QPointF XYSeries::at(int index)
+    \qmlmethod point XYSeries::at(int index)
     Returns the point at the position specified by \a index. Returns (0, 0) if
     the index is not valid.
 */
@@ -675,8 +691,16 @@ void QXYSeries::replace(int index, const QPointF &newPoint)
 void QXYSeries::replace(const QList<QPointF> &points)
 {
     Q_D(QXYSeries);
+    const bool needUpdateSelection = d->m_points.size() > points.size();
     d->m_points = points;
+    if (needUpdateSelection) {
+        d->m_selectedPoints.removeIf([maxIdx = points.size() - 1](int idx) {
+            return idx > maxIdx;
+        });
+    }
     emit pointsReplaced();
+    if (needUpdateSelection)
+        emit selectedPointsChanged();
 }
 
 /*!
@@ -775,7 +799,7 @@ void QXYSeries::clearPointsConfiguration(const QXYSeries::PointConfiguration key
     keys and QVariant values. For example:
     \code
     QLineSeries *series = new QLineSeries();
-    series->setName("Customized serie");
+    series->setName("Customized series");
     series->setPointsVisible(true);
 
     *series << QPointF(0, 6) << QPointF(2, 4) << QPointF(3, 6) << QPointF(7, 4) << QPointF(10, 5)
@@ -793,18 +817,22 @@ void QXYSeries::clearPointsConfiguration(const QXYSeries::PointConfiguration key
 
     series->setPointConfiguration(4, conf);
 
-    conf.remove(QXYSeries::PointConfiguration::LabelVisibility);
+    conf.remove(QXYSeries::PointConfiguration::Color);
+    conf[QXYSeries::PointConfiguration::LabelFormat] = "This Point";
     series->setPointConfiguration(6, conf);
     \endcode
 
-    In this example, you can see a default QLineSeries with 10 points
-    and with changed configuration of two points. Both changed points are
-    red and visibly bigger than the others with a look derived from series.
-    By default, points don't have labels, but the point at index 4 has
-    the label thanks to the QXYSeries::PointConfiguration::LabelVisibility
-    configuration value.
+    In this example, you can see a default QLineSeries with 10 points and with changed configuration
+    of two points. Both changed points are visibly bigger than the others with a look derived from
+    the series configuration.
+    By default, points don't have labels, but the point at index 4 has a label thanks to the
+    QXYSeries::PointConfiguration::LabelVisibility and QXYSeries::PointConfiguration::LabelFormat
+    configuration values.
+    The point at index 6 has a custom label \e {This Point} thanks to the
+    QXYSeries::PointConfiguration::LabelFormat configuration value.
     Below is an example of a chart created in this way:
-    \image xyseries_point_configuration.png
+    \image xyseries_point_configuration.png {Screenshot showing a line chart with two dots on it,
+           one being red and the other being blue}
 
     \sa pointsConfiguration(), clearPointsConfiguration()
     \since 6.2
@@ -833,24 +861,8 @@ void QXYSeries::setPointConfiguration(const int index, const QXYSeries::PointCon
                                       const QVariant &value)
 {
     Q_D(QXYSeries);
-    QHash<QXYSeries::PointConfiguration, QVariant> conf;
-    if (d->m_pointsConfiguration.contains(index))
-        conf = d->m_pointsConfiguration[index];
-
-    bool callSignal = false;
-    if (conf.contains(key)) {
-        if (conf[key] != value)
-            callSignal = true;
-    } else {
-        callSignal = true;
-    }
-
-    conf[key] = value;
-    d->m_pointsConfiguration[index] = conf;
-
-    if (callSignal) {
+    if (d->setPointConfiguration(index, key, value))
         emit pointsConfigurationChanged(d->m_pointsConfiguration);
-    }
 }
 
 /*!
@@ -921,16 +933,22 @@ void QXYSeries::sizeBy(const QList<qreal> &sourceData, const qreal minSize, cons
     }
 
     const qreal range = max - min;
+    const qreal sizeRange = maxSize - minSize;
+    bool changed = false;
 
     for (int i = 0; i < sourceData.size() && i < d->m_points.size(); ++i) {
         qreal pointSize = minSize;
         if (range != 0) {
             const qreal startValue = sourceData.at(i) - min;
             const qreal percentage = startValue / range;
-            pointSize = qMax(minSize, percentage * maxSize);
+            pointSize = minSize + (percentage * sizeRange);
         }
-        setPointConfiguration(i, QXYSeries::PointConfiguration::Size, pointSize);
+        if (d->setPointConfiguration(i, QXYSeries::PointConfiguration::Size, pointSize))
+            changed = true;
     }
+
+    if (changed)
+        emit pointsConfigurationChanged(d->m_pointsConfiguration);
 }
 
 /*!
@@ -1001,12 +1019,18 @@ void QXYSeries::colorBy(const QList<qreal> &sourceData, const QLinearGradient &g
     const qreal diff = min < 0 ? qAbs(min) : 0;
     min += diff;
 
+    bool changed = false;
+
     for (int i = 0; i < sourceData.size() && i < d->m_points.size(); ++i) {
         const qreal startValue = qMax(0.0, sourceData.at(i) + diff - min);
         const qreal percentage = startValue / range;
         QColor color = image.pixelColor(0, qMin(percentage * imgSize, imgSize - 1));
-        setPointConfiguration(i, QXYSeries::PointConfiguration::Color, color);
+        if (d->setPointConfiguration(i, QXYSeries::PointConfiguration::Color, color))
+            changed = true;
     }
+
+    if (changed)
+        emit pointsConfigurationChanged(d->m_pointsConfiguration);
 }
 
 /*!
@@ -1805,6 +1829,7 @@ void QXYSeriesPrivate::drawPointLabels(QPainter *painter, const QList<QPointF> &
             painter->setClipping(false);
 
         QList<int> pointsToSkip;
+        QHash<int, QString> labelFormats;
         QHash<int, int> offsets;
 
         if (!m_pointsConfiguration.isEmpty()) {
@@ -1812,14 +1837,16 @@ void QXYSeriesPrivate::drawPointLabels(QPainter *painter, const QList<QPointF> &
                 bool drawLabel = m_pointLabelsVisible;
                 if (m_pointsConfiguration.contains(i)) {
                     const auto &conf = m_pointsConfiguration[i];
-                    if (conf.contains(QXYSeries::PointConfiguration::LabelVisibility)) {
-                        drawLabel = m_pointsConfiguration
-                                            [i][QXYSeries::PointConfiguration::LabelVisibility]
-                                                    .toBool();
-
-                        if (drawLabel && conf.contains(QXYSeries::PointConfiguration::Size))
-                            offsets[i] = conf[QXYSeries::PointConfiguration::Size].toReal();
+                    auto key = QXYSeries::PointConfiguration::LabelVisibility;
+                    if (conf.contains(key)) {
+                        drawLabel = m_pointsConfiguration[i][key].toBool();
+                        key = QXYSeries::PointConfiguration::Size;
+                        if (drawLabel && conf.contains(key))
+                            offsets[i] = conf[key].toReal();
                     }
+                    key = QXYSeries::PointConfiguration::LabelFormat;
+                    if (conf.contains(key) && !conf[key].toString().isEmpty())
+                        labelFormats[i] = conf[key].toString();
                 }
 
                 if (!drawLabel)
@@ -1827,19 +1854,21 @@ void QXYSeriesPrivate::drawPointLabels(QPainter *painter, const QList<QPointF> &
             }
         }
 
-        drawSeriesPointLabels(painter, allPoints, offset, offsets, pointsToSkip);
+        drawSeriesPointLabels(painter, allPoints, offset, offsets, pointsToSkip, labelFormats);
     }
 }
 
 void QXYSeriesPrivate::drawSeriesPointLabels(QPainter *painter, const QList<QPointF> &points,
                                              const int offset, const QHash<int, int> &offsets,
-                                             const QList<int> &indexesToSkip)
+                                             const QList<int> &indexesToSkip,
+                                             const QHash<int, QString> &labelFormats)
 {
     if (points.size() == 0)
         return;
 
     static const QString xPointTag(QLatin1String("@xPoint"));
     static const QString yPointTag(QLatin1String("@yPoint"));
+    static const QString indexTag(QLatin1String("@index"));
 
     QFont f(m_pointLabelsFont);
     f.setPixelSize(QFontInfo(m_pointLabelsFont).pixelSize());
@@ -1853,9 +1882,10 @@ void QXYSeriesPrivate::drawSeriesPointLabels(QPainter *painter, const QList<QPoi
         if (indexesToSkip.contains(i))
             continue;
 
-        QString pointLabel = m_pointLabelsFormat;
+        QString pointLabel = labelFormats.contains(i) ? labelFormats[i] : m_pointLabelsFormat;
         pointLabel.replace(xPointTag, presenter()->numberToString(m_points.at(i).x()));
         pointLabel.replace(yPointTag, presenter()->numberToString(m_points.at(i).y()));
+        pointLabel.replace(indexTag, presenter()->numberToString(i));
 
         int currOffset = offset;
         if (offsets.contains(i))
@@ -1988,6 +2018,28 @@ QList<qreal> QXYSeriesPrivate::colorByData() const
     return m_colorByData;
 }
 
+bool QXYSeriesPrivate::setPointConfiguration(const int index,
+                                             const QXYSeries::PointConfiguration key,
+                                             const QVariant &value)
+{
+    QHash<QXYSeries::PointConfiguration, QVariant> conf;
+    if (m_pointsConfiguration.contains(index))
+        conf = m_pointsConfiguration[index];
+
+    bool changed = false;
+    if (conf.contains(key)) {
+        if (conf[key] != value)
+            changed = true;
+    } else {
+        changed = true;
+    }
+
+    conf[key] = value;
+    m_pointsConfiguration[index] = conf;
+
+    return changed;
+}
+
 QT_END_NAMESPACE
 
 #include "moc_qxyseries.cpp"
